/*{{{  process.c*/
#include <stdio.h>
#define STDIO_H
#include "adout.h"	/* for 'map_subscr_n()', etc */
#include "lex.h"
#include "output.h"	/* for 'earlier_error' */
#include "symbol.h"
#include "unit.h"	/* for 'this_unitp->scope', '->first_sp' */
#include "type.h"
#include "typedef.h"

/*{{{  < DISCUSSION -- what to do/check at end of decls >*/
#if 0
DISCUSSION -- what to do / check at end of decls
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
For each 'sp':

/*{{{  If 'sp->attrs' has 'AT_COMMON_BLOCK / _SAVE' set, and no other flags*/
If 'sp->attrs' has 'AT_COMMON_BLOCK / _SAVE' set, and no other flags
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

... and 'sp' has no type or array dims set, it could just be a common block
name, so we forget about it!

(If it *is* later used for a typed object, its type will be set as a
result of the reference).
/*}}}*/

/*{{{  Otherwise, if no (AT_ENTITY | AT_EXTERNAL) flag set, try to set one*/
Otherwise, if no (AT_ENTITY | AT_EXTERNAL) flag set, try to set one
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
if (  sp->arp  ||
     (sp->attrs & (AT_RESULT_VAR | AT_IN_COMMON | AT_EQUIV |
                   AT_INTENT | AT_SAVE | AT_TARGET | AT_IN_NAMELIST)) )

	sp->attrs |= AT_NORMAL_VAR;


So the entity type remains unknown only if 'sp->attrs' has only the
following flags set...

AT_DUMMY
AT_OPTIONAL
AT_EXTERNAL
AT_ACCESS

(N.B. We can't infer from the presence of a datatype that AT_NORMAL_VAR
should be set, as it might be an external / intrinsic function!).
/*}}}*/

/*{{{  Typing*/
Typing
~~~~~~
-- On o/p we add a type-dependent prefix to every ID, so we *must*
determine an ID's type, or absence of one, before it's first o/p.
Otherwise if its type is set *after* the first o/p, its name will change!

-- In the decls, we only o/p IDs that have appeared so far in the i/p,
so presumably each ID must have been declared, used, or is a dummy
argument.  Hence each 'sp' presumably *must* have 'sp->arp', or a
datatype, or some 'sp->attrs' flag set.

-- Hence, for all ID's *without a type* which are to be o/p during the
decls stage, 'process_decls()' must set an implicit type if ID could
**conceivably** be typed  [even if it really should be untyped!!].
We use the following rules:


(1) *Don't* set a data type if 'sp->attrs' has any of these flags set:

(AT_UNTYPED_ENTITY | AT_INFO)
	-- latter since the type will have been set if nec (e.g. for an  init
	   var / parameter, but *not* for an external / intrinsic procedure).


(2) *Set* a datatype if:

if (  sp->arp  ||
     (sp->attrs & (AT_ALLOC | AT_POINTER | AT_NORMAL_VAR | AT_STMT_VAR)) )


(N.B. AT_NORMAL_VAR subsumes lots of other attr flags which cause it to
be set in the previous check).


(3)  We have no idea of whether to set a data type if the *only* flags
in 'sp->attrs' are:

AT_DUMMY	-- var or dummy procedure?
AT_EXTERNAL	-- func or subr?
AT_OPTIONAL
AT_ACCESS

/*{{{  What to do if only flags are AT_DUMMY and/or AT_EXTERNAL ?*/
What to do if only flags are AT_DUMMY and/or AT_EXTERNAL
--------------------------------------------------------

-- If AT_EXTERNAL is set (with or without AT_DUMMY), o/p an EXTERNAL
decl *without* a datatype (even if one is set), since it will be
converted to a subroutine call.  Also use the untyped prefix!

*****************************************************************
In this case we defer giving it an implicit datatype, since we know
that the o/p uses the untyped prefix.  An implicit datatype
can be installed if and when 'sp' is referenced as a function.
*****************************************************************


-- If AT_DUMMY is set *without* AT_EXTERNAL (or AT_INFO, which implies
an explicit interface), then it's either a dummy var (which must be typed)
or a dummy function or subroutine (but a local name, *not* an external one).

*****************************************************************
In this case it *is* o/p among the decls (viz in the dummy argument
list) so we must install a 'tentative' datatype and use that to
determine its name for o/p.
	There's just one proviso -- if IMPLICIT NONE is effective,
then *don't* try to install a type -- we assume that it really
should be untyped, and therefore is a dummy subroutine.
*****************************************************************

However, we don't actually o/p a declaration for the name.  If it's a
dummy procedure, since there's no EXTERNAL decl we needn't output one.
Likewise, if it's a dummy var we needn't declare it -- the implicit
type rules will apply.

We must be careful to spot this case in 'stmt (%n)', when generating
the prefix for such an ID.  Viz, we must do the following check:

if (sp->attrs & AT_PROCEDURE) {
	if (  (sp->attrs & AT_DUMMY)  &&
            ! (sp->attrs & (AT_INFO | AT_EXTERNAL)) )

		-- o/p prefix according to installed 'sp->type'

	else if ( ! (sp->attrs & AT_INFO)  ||
	          ! (sp->info.sp->type & TY_ELEMENTAL) )

		-- o/p untyped prefix
	else
		-- it's an elemental procedure, so don't add
		   *any* prefix!
}	

N.B. AT_SUBR can have a type flag installed and it doesn't matter
-- it won't affect the o/p.


/*}}}*/

/*{{{  Special considerations typing in an interface block*/
Special considerations typing in an interface block
---------------------------------------------------

For a dummy arg 'sp' without a datatype set, and with only flags
(AT_DUMMY | AT_EXTERNAL) set, we **do** set a (tentative) datatype,
however, in case it's a dummy variable or dummy function.

We do this because outside the interface block we don't know the
implicit typing rules in effect therein, so we can't give the dummy
args an implicit type later.

This means that a dummy subroutine in a procedure interface
may have an implicit datatype -- but it doesn't really matter!

/*}}}*/
/*}}}*/

/*{{{  Special considerations typing in an interface block*/
/*}}}*/

/*{{{  Other checks*/
Other checks
~~~~~~~~~~~~
-- global, non-pointer arrays must have const bounds

if (sp->arp  &&  (sp->arp->attrs & DEFERRED_SHAPE)) {
	if (! (AT_ALLOC | AT_POINTER)) {
		if (AT_DUMMY)
			DEFERRED_SHAPE -> ASSUMED_SHAPE
		else {
			earlier_error = TRUE;
			errmsg ("array %q must have explicit dimensions");
			earlier_error = FALSE;
		}
	}
}

-- In 'arp->attrs', could also install a flag indicating whether
dimension appeared in or out of an HPF directive.  Then, give
errmsg if procs/template dimensioned *outside* HPF dir, or if
other object dimensioned *in* HPF dir.
	These errors are already checked if identity of 'sp'
known at time of dimensioning  (in which case 'sp' *isn't*
given array dimensions -- so no special problems for the check
proposed here).  This check will trap cases where dims installed
before *any* other info is given about 'sp'. E.g.:

      dimension t(10)
!hpf$ template t    ! dim info should be in HPF dir

!hpf$ dimension :: x(20)
      real x        ! dim info should be out of HPF dir
/*}}}*/

#endif

/*}}}*/

/******************************************************
WHAT ARE REAL DIFFERENCES BETWEEN NEXT 2 (IF ANY).
COULD WE COMBINE THEM INTO ONE 'prcess_decls()' ??
*******************************************************/
/*{{{  process_proc_interface () -- this also calls 'process_decls()'*/
process_proc_interface ()
/*---------------------*/
/*--------------------------------------------------------------*/
/*	Called by 'close_current_unit ()' when we're at the end	*/
/* of a procedure interface body, to ensure that the		*/
/* information in the procedure's symtab is as complete as	*/
/* possible.							*/
/*--------------------------------------------------------------*/

/*********************
-- How do we allow for sp's in this scope unit before 'proc_sp'
	(e.g. type-name, extrinsic-model-name) ?
-- If separate result-var ('result_sp'), 'proc_sp->mapp' doesn't
	yet have 'result_sp->mapp' attached!
************************/
{
	Symbol  *proc_sp,  *sp;

	proc_sp = this_unitp->sp;	/** N.B. WHAT IF THERE ARE 'sp's
					    BEFORE THE proc_name (E.G.
					    type-name OR extrinsic-name)? **/

	if (proc_sp->attrs & AT_RESULT_VAR) {
		/*{{{  install datatype & entity type for function name.*/
		/*----------------------------------------------------------------
			If the function result variable has no datatype set,
		install an implicit one.
			Also, the only possible entity types for a result
		variable are (AT_NORMAL_VAR | AT_POINTER).  If neither are set,
		install AT_NORMAL_VAR.
		
		N.B.  If AT_FUNC, rather than AT_RESULT_VAR, is set in 'proc_sp',
		then there's a separate result variable.  In that case 'proc_sp'
		should 'inherit' its type from the result variable, if not
		explicitly set in the 'function-stmt', which is done at the end
		of this funciton.
		----------------------------------------------------------------*/
		lookup_or_install_type (proc_sp);
		if ( ! (proc_sp->attrs & AT_ENTITY))
			proc_sp->attrs |= AT_NORMAL_VAR;
		/*}}}*/
	}

	for (sp = proc_sp->nextid;  sp  &&  (sp->attrs & AT_DUMMY);
	                                                 sp = sp->nextid) {
		if (*(sp->name) == '*')
			continue;

		if ( ! (sp->attrs & AT_ENTITY) ) {
			/*{{{  try to deduce and set 'sp's entity type*/
			/*----------------------------------------------------------------
				'sp->attrs' has no AT_ENTITY flag set, so try to
			deduce 'sp's entity type.  I.e., if it has INTENT or TARGET
			declared, or is in a namelist group, set AT_NORMAL_VAR,
			while if it's declared EXTERNAL and has a datatype or
			dimensions, set AT_FUNC.
			
			N.B.  A dummy argument may only have the following flags set in
			'sp->attrs':
			
				(AT_POINTER | AT_NORMAL_VAR | AT_FUNC | AT_SUBR)  -- AT_ENTITY flags
			
				AT_EXTERNAL
			
				AT_IN | AT_OUT
				AT_TARGET
				AT_IN_NAMELIST
					
				AT_INFO (e.g. for dummy proc with expl interface)
			
				AT_DUMMY			-- always set
				(AT_COMMON_BLOCK | AT_COMMON_SAVE | AT_OPTIONAL)
								-- but these are irrelevant for
								-- determining the entity type.
			----------------------------------------------------------------*/
			
			if ( ! (sp->attrs & AT_EXTERNAL)) {
				if (sp->arp  || (sp->attrs & (AT_INTENT | AT_TARGET | AT_IN_NAMELIST)) )
			
					sp->attrs |= AT_NORMAL_VAR;
			}
			
			/*--------------------------------------------------------------*/
			/* ... else 'sp' has AT_EXTERNAL set but not (AT_FUNC |AT_SUBR).*/
			/* If it has a type or dimensions, it must be AT_FUNC.		*/
			/*--------------------------------------------------------------*/
			else if ((sp->type & TY_TYPE) || sp->arp) {
				sp->attrs |= AT_FUNC;
				lookup_or_install_type (sp);	/* in case shape but
								   not type declared */
			}
			/*}}}*/
		}

		/*{{{  if no datatype, set it if 'sp' could conceivably be typed*/
		/*----------------------------------------------------------------
			If a datatype isn't specified for 'sp', install its
		implicit type if 'sp' could *conceivably* be a typed entity,
		i.e. if AT_SUBR isn't set.
			In some cases this datatype is installed when it shouldn't
		be, as 'sp' might be a dummy subroutine which can't be recognised
		as such from the given decls.  (A dummy subroutine is only
		specified as such if an explicit interface is given for it;
		otherwise, the absence of a typedecl and/or the specification of
		an EXTERNAL attribute is insufficient to resolve the issue!)
		If the first use of 'proc_sp' shows that dummy argument 'sp'
		is indeed a dummy subroutine, then we must 0 any implicit type
		that may have been set.
			We install a datatype speculatively now, rather than
		waiting until we know for sure whether dummy argument 'sp' is a
		typed entity (i.e. from a reference to the procedure), as it'll
		be too late then, since this interface body may have it's own
		implicit type mapping which is different from that used elsewhere!
		
		N.B.  We also don't install a datatype if AT_FUNC is set, since:
		
		-- if 'sp' has an explicit interface, then the type is taken from
		that  (i.e. from 'sp->info.sp');
		
		-- in all other circumstances in which AT_FUNC can be set
		(i.e. via a reference to 'sp', or as a result of the earlier
		analysis in this function) the datatype is set at the same time
		as AT_FUNC.
		----------------------------------------------------------------*/
		if ( ! (sp->attrs & (AT_SUBR | AT_FUNC)) )
			lookup_or_install_type (sp);
		/*}}}*/

		/*{{{  do checking & give any necessary errmsgs*/
		/*--------------------------------------------------------------*/
		/* Check dummy arg decls and give any necessary errmsgs		*/
		/* (e.g. should a dummy function have an explicit interface?)	*/
		/*--------------------------------------------------------------*/
		if ((sp->attrs & AT_FUNC)  &&  !(sp->attrs & AT_INFO)) {
			if (sp->arp)
				errmsg ("array function %q must have an explicit interface",
				         sp->name);
		
			else if (   (sp->type & TY_CHARACTER)  &&  sp->typeinfo.char_len  &&
			          ! (sp->typeinfo.char_len->type & EX_INT_LIT_CONST)  &&
			            (sp->typeinfo.char_len->token != ND_STAR) )
		
				errmsg ("function %q with variable length character result must have an explicit interface",
				         sp->name);
		}
		/*}}}*/
	}

	if (sp  &&  (sp->attrs & AT_RESULT_VAR)) {
		/*{{{  separate result var; install type & copy to 'proc_sp'*/
		/*----------------------------------------------------------------
			'sp' is a separate result variable;  'proc_sp->attrs' has
		AT_FUNC, rather than AT_RESULT_VAR, set.
			In this case do the same as if 'proc_sp' were the result
		variable, namely:  install an implicit datatype in 'sp->type'
		if necessary, and install AT_NORMAL_VAR in 'sp->attrs' if no
		entity type  (i.e. AT_NORMAL_VAR or AT_POINTER)  is set.
			Also, copy all of this information to 'proc_sp',
		so that outside the interface 'proc_sp' has a copy of all the
		result variable's properties.
		
		N.B.  In this case, decls after the 'func-stmt' must *not* refer
		to the 'func-name', but instead must refer to the 'result-var'.
		We have ensured that this is enforced -- any attempt to install an
		attribute, type or shape in 'proc_sp' results in an errmsg if:
			((proc_sp->attrs & AT_FUNC)  && (proc_sp == this_unitp->sp))
		----------------------------------------------------------------*/
		lookup_or_install_type (sp);
		if ( ! (sp->attrs & AT_ENTITY))
			sp->attrs |= AT_NORMAL_VAR;
		
		
		/*--------------------------------------------------------------*/
		/* Copy info to 'proc_sp'.					*/
		/*								*/
		/* N.B. It's OK to overwrite info already in 'proc_sp', as it	*/
		/* could only have been installed during the 'func_stmt' and	*/
		/* the same info would have been simultaneously installed in	*/
		/* the result var (i.e. this 'sp').				*/
		/*--------------------------------------------------------------*/
		proc_sp->attrs = sp->attrs;	/* N.B. 'proc_sp' becomes 'AT_RESULT_VAR'
						   rather than 'AT_FUNC' */
		proc_sp->type  = sp->type;
		if (sp->type & TY_DERIVED)
			proc_sp->typeinfo.type_sp = sp->typeinfo.type_sp;
		
		else if ((sp->type & TY_CHARACTER)  &&  sp->typeinfo.char_len  &&
		                                 ! proc_sp->typeinfo.char_len)
			attach_expr (proc_sp->typeinfo.char_len,  sp->typeinfo.char_len);
		
		if (sp->arp)
			attach_array (proc_sp->arp, sp->arp);
		/*}}}*/
	}
	process_decls ();
}
/*}}}*/
/*{{{  process_decls () -- ok*/
process_decls ()
/*------------*/
/*--------------------------------------------------------------*/
/*	This is called by:					*/
/* -- 'set_part_and_label()', after the 1st executable stmt in	*/
/*	a program unit or subprogram is seen;			*/
/* -- 'close_current_unit ()', if an external program unit is	*/
/*	'END'ed *without* any executable stmts!			*/
/*--------------------------------------------------------------*/
{
	bool   must_compose_align = FALSE;
	short  this_scope = this_unitp->scope;
	Symbol *sp;
	Array  *arp;


	earlier_error = TRUE;		/* for benefit of any errmsgs */
	for (sp = this_unitp->first_sp;  sp;  sp = sp->nextid) {
		if ( (sp->scope != this_scope)  ||
		     (*(sp->name) == '*') )
			continue;

		/*{{{  try to deduce 'sp's 'entity type' & datatype if necessary*/
		if (    only_flags_set (sp->attrs, AT_COMMON_NAME)  &&
		     ! (sp->type & TY_TYPE)  &&  ! sp->arp )
		
			goto  done_type_checks;	/* assume it's just a COMMON block name! */
		
		/*--------------------------------------------------------------*/
		/* If 'sp->attrs' has no (AT_ENTITY | AT_EXTERNAL) flag set,	*/
		/* set 'AT_NORMAL_VAR' if it can be deduced.			*/
		/*								*/
		/* N.B.  The entity type remains unknown if the only flags set	*/
		/* in 'sp->attrs' are:  (AT_DUMMY | AT_OPTIONAL | AT_EXTERNAL |	*/
		/* AT_ACCESS | AT_COMMON_NAME)					*/
		/*--------------------------------------------------------------*/
		if ( ! (sp->attrs & (AT_ENTITY | AT_EXTERNAL))  &&
		     (  sp->arp  ||
		       (sp->attrs & (AT_RESULT_VAR | AT_IN_COMMON | AT_EQUIV |
		                     AT_INTENT | AT_SAVE | AT_TARGET | AT_IN_NAMELIST)) ) )
		
			sp->attrs |= AT_NORMAL_VAR;
		
		
		/*--------------------------------------------------------------*/
		/* If 'sp' is a data object, install an implicit datatype if	*/
		/* necessary.							*/
		/*								*/
		/* N.B. We don't know whether to install a datatype if the only	*/
		/* flags set in 'sp->attrs' are:   (AT_DUMMY | AT_EXTERNAL |	*/
		/* AT_OPTIONAL | AT_ACCESS | AT_COMMON_NAME)			*/
		/*--------------------------------------------------------------*/
		
		if ((sp->type & TY_TYPE)  ||
		    (sp->attrs & (AT_UNTYPED_ENTITY | AT_INFO | AT_EXTERNAL)) )
			/*------------------------------------------------------*/
			/* ...already typed, or a non-data entity, or the	*/
			/* typing decision can be deferred (for EXTERNAL), so	*/
			/* do nothing!						*/
			/*------------------------------------------------------*/
			;
		
		else if ( sp->arp  ||
		         (sp->attrs & (AT_ALLOC | AT_POINTER | AT_NORMAL_VAR | AT_STMT_VAR)) ||
		         ( (sp->attrs & AT_DUMMY)  &&
		                ! (this_unitp->impl_type [*sp->name - 'A'] & TY_NONE)) )
			/*------------------------------------------------------*/
			/* N.B. In the case of 'AT_DUMMY', we install a		*/
			/* 'tentative' data type unless 'IMPLICIT NONE' has	*/
			/* been declared.					*/
			/* (See the 'Description' at the start of this file.)	*/
			/*------------------------------------------------------*/
			lookup_or_install_type (sp);
		
		done_type_checks :
			;
		/*}}}*/

		if (sp->arp  &&  ! (sp->arp->attrs & EXPL_SHAPE)) {
			/*{{{  for assumed/deferred shape/size, supply missing 'l'&'n'*/
			/*--------------------------------------------------------------*/
			/* Some 'n', and possibly also 'l', are omitted (unless		*/
			/* (nlinks > 1) and this 'arp' has already been processed!).	*/
			/* Supply values.						*/
			/*--------------------------------------------------------------*/
			char  *name = sp->name;
			int    dim,  rank = (sp->arp->attrs & RANK_MASK);
			Expr   **lp = sp->arp->l,
			       **np = sp->arp->n;
			
			
			for (dim = 0;   dim < rank;   ++dim, ++lp, ++np) {
				if ( ! *lp)
					attach_expr (*lp, int_lit_1);
			
				if ( ! *np  ||  (*np)->token == ND_STAR) {
					/*------------------------------------------------
					N.B. This 'arp->n' is useless for dummy args in an
					explicit interface!  It's just provided so that
					'install_id_map ()' doesn't crash.  Hence the latter
					installs a flag 'MA_ASSUMED_SHAPE' in 'mapp->attrs'
					to signal that 'mapp's entries should be ignored in
					thsi case.
					------------------------------------------------*/
					int  elem = map_subscr_n (rank, dim);
			
					if (*np)
						delete_expr (*np, DECR_NLINKS);
					attach_expr (*np, dummy_map_primary (name, elem));
				}
			}
			/*}}}*/
		}

		/*{{{  finish checks for HPF mapping*/
		/*----------------------------------------------------------------
		Finish checks for aligned or distributed objects.
		
		N.B. For alignment, we check the align target before the alignee
		in case the same 'alignp' is shared by several alignees.
			In that case, if there are problems with the target, the
		record is corrected and applies to *all* alignees.  However, if
		the alignee is checked first and there are problems with it,
		a new corrected record is made just for this alignee.  Then if a
		problem is also found with the target it must be corrected in
		>= 2 'alignp' records, not just one, causing duplication of errmsgs.
			Likewise, for distribution we check the target before
		the distributee for similar reasons.
		----------------------------------------------------------------*/
		if (sp->alignp) {
			/*------------------------------*/
			/* i.e. "ALIGN" or "INHERIT"!	*/
			/*------------------------------*/
			Align   *alignp = sp->alignp;
			Symbol  *tgt_sp = alignp->tgt_sp;	/* NULL for 'INHERIT' */
		
			if ( ! (alignp->attrs & AL_CHECKED_TGT))
				check_align_target (alignp, sp);
		
			if ( ! sp->arp)
				check_alignee (sp);
		
			if (tgt_sp  &&  tgt_sp->alignp)
				must_compose_align = TRUE;
		}
		else if (sp->distp) {
			Distr   *distp   = sp->distp;
			Symbol  *proc_sp = distp->proc_sp;	/* NULL for 'INHERIT_PROCS'*/
		
			if (proc_sp) {
				if ( ! (distp->attrs & DI_CHECKED_PROC_TYPE)) {
					if ( ! (proc_sp->attrs & AT_PROCESSORS))
						errmsg ("DISTRIBUTE target %q is not a processors arrangement",
						         proc_sp->name);
					
					distp->attrs |= DI_CHECKED_PROC_TYPE;
				}
				if ( ! (distp->attrs & DI_CHECKED_PROC_DIMS))
					check_distr_target (distp, sp);
			}
			if ( ! sp->arp)
				check_distributee (sp);
		
			if (distp->attrs & DI_NEED_BLOCKSIZE_MASK)
				calc_default_blocksizes (sp);
		}
		/*}}}*/
	}

	if (must_compose_align) {
		/*{{{  pass through the symtab, combining multiple alignments*/
		/*--------------------------------------------------------------*/
		/* Pass through the symtab, combining all multiple alignments	*/
		/* to single ones.						*/
		/*--------------------------------------------------------------*/
		for (sp = this_unitp->first_sp;  sp;  sp = sp->nextid) {
			if (sp->scope != this_scope)
				continue;
		
			if (sp->alignp  &&  sp->alignp->tgt_sp  &&
			    sp->alignp->tgt_sp->alignp)
				compose_align (sp, sp, 1);
		}
		/*}}}*/
	}

	/*{{{  pass through the symtab, installing 'sp->mapp' where appropriate*/
	/*----------------------------------------------------------------
		Pass through the symtab, setting up a 'Map' structure,
	'sp->mapp', for	each ID 'sp' that requires one, viz:
	
	-- all data arrays (i.e. excluding PROCESSOR and TEMPLATE arrays),
		even if they're not aligned or distributed.
	
	-- scalars if and only if they're explicitly distributed, inherited,
		or aligned with an ultimate align target that is explicitly
		distributed or inherited.
			(I.e. *don't* assign a 'mapp' to a globally
		replicated scalar.  If a scalar's ultimate align target
		isn't explicitly distributed or inherited, the scalar
		is assumed to be globally replicated.)
	
	N.B.
	-- It isn't necessary to assign a 'mapp' to an ID simply because
		it's a dummy argument!  (On entry, dummy scalars are
		assumed to be globally replicated, and dummy arrays
		are assumed to have the INHERIT attribute -- i.e. any
		legal mapping -- unless these defaults are over-ridden
		by descriptive mapping directives.).
	
	-- if a data entity 'sp' has no 'sp->mapp', then it's a replicated
	scalar!
	----------------------------------------------------------------*/
	for (sp = this_unitp->first_sp;  sp;  sp = sp->nextid) {
		if ( (sp->scope != this_scope)  ||
		     (sp->attrs & (AT_TEMPLATE | AT_PROCESSORS)) )
			continue;
	
		if (  sp->arp  ||			/* an array, distrib or not!*/
		      sp->distp  ||			/* a distributed scalar */
		     (sp->alignp  &&
		       (! sp->alignp->tgt_sp  ||	/* an INHERITed scalar */
		          sp->alignp->tgt_sp->distp ||	/* has distrib align tgt */ 
		          sp->alignp->tgt_sp->alignp))	/* has INHERITed align tgt*/
		   ) 
	
			install_id_map (sp);
	}
	/*}}}*/

	earlier_error = FALSE;
	
/*{{{  *** PREVIOUS STUFF !! ****/
#if 0

int  type  = sp->type & TY_TYPE;
int  object = sp->attrs & (AT_ENTITY | AT_EXTERNAL | AT_INTRINSIC);

/*----------------------------------------------*/
/*  Now know its a datatype object or EXTERNAL	*/
/*  (i.e. a FN or SUBR).  Try to narrow-down	*/
/*  any uncertainty about the object's identity	*/
/*  that may still exist:			*/
/*----------------------------------------------*/

if ((object & AT_EXTERNAL)  &&  (sp->type & TY_TYPE)) {
	clear_flags (sp->attrs, AT_EXTERNAL);
	sp->attrs |= AT_FUNC;
}

#endif
/*}}}*/
}
/*}}}*/


/*{{{  process_callstmt (Symbol *proc)*/
process_callstmt (proc)
/*-------------------*/
	Symbol *proc;
{
	bool  ok = TRUE;
	Symbol *dummy,  *actual;

	/*--------------------------------------*/
	/*  check actual arguments of CALL stmt	*/
	/*--------------------------------------*/
	for ( dummy = proc->nextid;
	      ok  &&  dummy  &&  (dummy->attrs & AT_DUMMY);
	      dummy = dummy->nextid ) {

		actual = dummy->info.sp;
		ok = argmatch (proc, actual, dummy);
	}

	/*--------------------------------------*/
	/*  write CALL stmt, and remove links	*/
	/*  to actual arguments.		*/
	/*--------------------------------------*/
	if (ok) {
		printf ("CALL %s (", proc->name);
		dummy = proc->nextid;
		if (dummy  &&  (dummy->attrs & AT_DUMMY) ) {

			actual = dummy->info.sp;
			printf ("%s", actual->name);
			dummy->info.sp = NULL;
		}
		for ( dummy = dummy->nextid;
		      dummy  &&  (dummy->attrs & AT_DUMMY);
		      dummy = dummy->nextid ) {

			actual = dummy->info.sp;
			printf (", %s", actual->name);
			dummy->info.sp = NULL;
		}
		printf (")\n");
	}
	else
		puts (textbuf);

}
/*}}}*/

/*}}}*/
